// Copyright 2017-2019, Nicholas Sharp and the Polyscope contributors. http://polyscope.run.
#pragma once

enum class ShadeStyle { FLAT = 0, SMOOTH };
enum class MeshElement { VERTEX = 0, FACE, EDGE, HALFEDGE, CORNER };
