// Copyright 2017-2019, Nicholas Sharp and the Polyscope contributors. http://polyscope.run.
#pragma once

#include <string>


namespace polyscope {

std::string promptForFilename(std::string filename = "out");
}
