// Copyright 2017-2019, Nicholas Sharp and the Polyscope contributors. http://polyscope.run.

#include <iostream>
#include <stdexcept>

using std::cout;
using std::endl;

namespace polyscope {
namespace render {

} // namespace gl
} // namespace polyscope
